"""
This module contains functions for a direct simulation
of gravitational 2-body system using realistic data for various
initial configurations
"""

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from sunpy.coordinates import (get_body_heliographic_stonyhurst, get_horizons_coord)
from sunpy.time import parse_time

from jobs.src.bht_algorithm_3D import MainApp, Particle
from jobs.src.integrators import verlet
from jobs.src.system import GravitationalSystem
from tasks.src.utils import (coord_to_polar, xyz_to_polar)

# system settings
M_S = 1  # solar mass
M_E = 3.00e-6 * M_S
M_L = 3.69e-8 * M_S
G = 4 * np.pi**2  # AU^3 / m / yr^2

# JPL data
time = parse_time('2024-01-01 00:00')

moon = get_horizons_coord('301', {
    'start': time - 40 * u.day,
    'stop': time + 40 * u.day,
    'step': '1d'
},
                          include_velocity=True)

earth = get_horizons_coord('399', {
    'start': time - 40 * u.day,
    'stop': time + 40 * u.day,
    'step': '1d'
},
                           include_velocity=True)

sun = get_body_heliographic_stonyhurst('Sun', time)  # (lon, lat, radius) in (deg, deg, AU)
moon = moon.transform_to(sun)
earth = earth.transform_to(sun)

# simulation settings
T = 0.23  # yr
n_order = 6
dt = 4**(-n_order)  # yr


# force computation
def force(q: np.ndarray) -> np.ndarray:
    r1 = q[0:3]
    r2 = q[3:6]

    return np.array([
        -G * M_E * (M_S * r1 / np.linalg.norm(r1)**3 + M_L * (r1 - r2) / np.linalg.norm(r1 - r2)**3),
        -G * M_L * (M_S * r2 / np.linalg.norm(r2)**3 + M_E * (r2 - r1) / np.linalg.norm(r2 - r1)**3),
    ]).flatten()


# run direct simulation
moon_0_pos = moon[0].represent_as('cartesian')
earth_0_pos = earth[0].represent_as('cartesian')

x0 = np.array([
    earth_0_pos.x.value, earth_0_pos.y.value, earth_0_pos.z.value, moon_0_pos.x.value, moon_0_pos.y.value,
    moon_0_pos.z.value, earth[0].velocity.d_x.value, earth[0].velocity.d_y.value, earth[0].velocity.d_z.value,
    moon[0].velocity.d_x.value, moon[0].velocity.d_y.value, moon[0].velocity.d_z.value
])

system = GravitationalSystem(r0=x0[:6],
                             v0=x0[6:],
                             m=np.array([M_E, M_E, M_E, M_L, M_L, M_L]),
                             t=np.linspace(0, T, int(T // dt)),
                             force=force,
                             solver=verlet)

t, p, q = system.simulation()


# run bht simulation
# force computation
def force(q: np.ndarray) -> np.ndarray:
    r1 = q[0:3]  # Earth coordinates
    r2 = q[3:6]  # Moon coordinates

    sun = Particle(0, 0, 0, M_S)
    earth = Particle(r1[0], r1[1], r1[2], M_E)
    moon = Particle(r2[0], r2[1], r2[2], M_L)

    particles = [sun, earth, moon]

    barnes_hut = MainApp()  # Initialize Barnes-Hut algorithm instance
    barnes_hut.BuildTree(particles)  # Build the Barnes-Hut tree with particles
    barnes_hut.rootNode.ComputeMassDistribution()  #Compute the center of mass of the tree nodes

    f_earth = barnes_hut.rootNode.CalculateForceFromTree(earth)
    f_moon = barnes_hut.rootNode.CalculateForceFromTree(moon)

    return np.concatenate((f_earth, f_moon), axis=0)


system = GravitationalSystem(r0=x0[:6],
                             v0=x0[6:],
                             m=np.array([M_E, M_E, M_E, M_L, M_L, M_L]),
                             t=np.linspace(0, T, int(T // dt)),
                             force=force,
                             solver=verlet)

t, p_bht, q_bht = system.simulation()

# plot trajectory in polar coordinates
q_E_polar = np.array([xyz_to_polar(pos[0], pos[1], pos[2]) for pos in q])
q_M_polar = np.array([xyz_to_polar(pos[3], pos[4], pos[5]) for pos in q])
q_E_bht_polar = np.array([xyz_to_polar(pos[0], pos[1], pos[2]) for pos in q_bht])
q_M_bht_polar = np.array([xyz_to_polar(pos[3], pos[4], pos[5]) for pos in q_bht])

fig = plt.figure()
ax = fig.add_subplot(projection='polar')
ax.plot(0, 0, 'o', label='Sun', color='pink', markersize=10)
ax.plot(q_E_polar[:, 0], q_E_polar[:, 1], label='Earth', color='blue', linestyle='-.')
ax.plot(q_M_polar[:, 0], q_M_polar[:, 1], label='Moon', color='purple', linestyle='dashed')
ax.plot(q_E_bht_polar[:, 0], q_E_bht_polar[:, 1], label='Earth(BHT)', linestyle='-.')
ax.plot(q_M_bht_polar[:, 0], q_M_bht_polar[:, 1], label='Moon(BHT)', linestyle='dashed')
ax.plot(*coord_to_polar(earth.transform_to(sun)), label='Earth(JPL)', color='green', linestyle='-.')
ax.plot(*coord_to_polar(moon.transform_to(sun)), label='Moon(JPL)', color='red', linestyle='dashed')
ax.set_title('Stonyhurst heliographic coordinates')
ax.legend(loc='upper left', bbox_to_anchor=(1.0, 0.5))
plt.show()

# plot trajectory in Cartesian coordinates
moon_xyz = moon.represent_as('cartesian')
earth_xyz = earth.represent_as('cartesian')
moon_JPL = np.array([(i.x.value, i.y.value, i.z.value) for i in moon_xyz])
earth_JPL = np.array([(i.x.value, i.y.value, i.z.value) for i in earth_xyz])

fig = plt.figure()
ax = fig.add_subplot(projection='3d')
ax.plot(0, 0, 0, 'o', label='Sun', color='pink', markersize=10)
ax.plot(q[:, 0], q[:, 1], q[:, 2], label="Earth")
ax.plot(q[:, 3], q[:, 4], q[:, 5], label="Moon")
ax.plot(q_bht[:, 0], q_bht[:, 1], q_bht[:, 2], label="Earth(BHT)")
ax.plot(q_bht[:, 3], q_bht[:, 4], q_bht[:, 5], label="Moon(BHT)")
ax.plot(earth_JPL[:, 0], earth_JPL[:, 1], earth_JPL[:, 2], '--', label='Earth(JPL)')
ax.plot(moon_JPL[:, 0], moon_JPL[:, 1], moon_JPL[:, 2], '-.', label='Moon(JPL)')
ax.set_xlabel("X")
ax.set_ylabel("Y")
ax.set_zlabel("Z")
ax.set_xticks([])
ax.set_yticks([])
ax.set_zticks([])
ax.legend()
plt.show()
